class PrototypeCoop expands CoopGame;

var(General) config bool bAllowNoBlock;
var(General) config bool bAllowVoteSkip;
var(General) config bool bCustomFriendlyFireImpact;
var(General) config bool bKeepScores;
var(General) config bool bNoFriendlyFireImpact;
var(General) config bool bNoInventoryRespawns;
var(General) config bool bNoTeamKills;
var(General) config bool bPlayersCanViewAnyActors;
var(General) config bool bRespawnPickups;
var(General) config bool bRetainInventory;
var(General) config bool bShowCoopKillsCount;
var(General) config bool bShowCreatureHealth;
var(General) config bool bShowCreatureNames;
var(General) config bool bShowDamageFromPlayers;
var(General) config bool bShowEnds;
var(General) config bool bShowPlayerKills;
var(General) config float ExtraServerTravelPause;
var(General) config int IdleTimeThreshold;
var(General) config int MaxAddedNaliFruits;
var(General) config float PickupRespawnTime;
var(General) config float PlayerRespawnDelay;
var(General) config int VoteEndMinPct;
var(General) config int VoteSkipMinPct;
var(General) config int VoteSkipMinPlayers;
var(General) config float WeaponRespawnTime;

var UGoldMP GameControl;

struct PlayerScoresInfo
{
	var string PlayerName;
	var string IPAddress;
	var int Score;
	var int Deaths;
};

var PlayerScoresInfo SavedScores[256];

event InitGame(string Options, out string Error)
{
	InitGameControl();
	SaveConfig();
	Super.InitGame(Options, Error);
	GameControl.AdminControl.InitGame(Options, Error);
}

function InitGameControl()
{
	GameControl = class'UGoldMP'.static.InitGameControl(self);
	if (GameControl == none)
		Error("Failed to spawn an instance of GameControl");
}

event PreLogin(string Options, out string Error)
{
	if (GameControl.VerifyPreLogin(Options, Error))
		super.PreLogin(Options, Error);
}

event PlayerPawn Login(
	string Portal,
	string Options,
	out string Error,
	class<PlayerPawn> SpawnClass)
{
	if (NumPlayers + NumSpectators >= 20)
	{
		Error = "Too many players on the server";
		return none;
	}
	GameControl.AdjustLoginParams(Portal, Options, Error, SpawnClass);
	return GameControl.InitLoggedInPlayer(GameControl.AdminControl.UserLogin(Portal, Options, Error, SpawnClass), Options);
}

event PostLogin(PlayerPawn NewPlayer)
{
	Super.PostLogin(NewPlayer);
	GameControl.PostLogin(NewPlayer);
	if (bKeepScores)
		ObtainSavedScores(GameControl.GetPlayerInfo(NewPlayer));
}

function Logout(Pawn Exiting)
{
	if (ScriptedPawn(Exiting) == none && !Exiting.IsA('SpaceMarine'))
		GameControl.AdminControl.PlayerLogout(Exiting);
	if (bKeepScores)
		UpdateSavedScores(GameControl.GetPlayerInfo(Exiting));
}

function NavigationPoint FindPlayerStart(byte Team, optional string IncomingName)
{
	local NavigationPoint Candidate[32];
	local NavigationPoint Best;
	local PlayerStart PStart;
	local Teleporter Telep, OtherTelep;
	local string TelepTag;
	local float Score[32], BestScore, NextDist;
	local Pawn OtherPlayer;
	local int i, num;

	num = 0;

	foreach AllActors(class'PlayerStart', PStart)
		if (PStart.bCoopStart && PStart.bEnabled && !PStart.Region.Zone.bWaterZone && PStart.Region.ZoneNumber > 0)
		{
			if (num < ArrayCount(Candidate))
				Candidate[num] = PStart;
			else if (Rand(num) < ArrayCount(Candidate))
				Candidate[Rand(ArrayCount(Candidate))] = PStart;
			++num;
		}

	if (num == 0)
		foreach AllActors(class'PlayerStart', PStart)
		{
			if (PStart.bCoopStart && PStart.bEnabled && PStart.Region.ZoneNumber > 0)
			{
				if (num < ArrayCount(Candidate))
					Candidate[num] = PStart;
				else if (Rand(num) < ArrayCount(Candidate))
					Candidate[Rand(ArrayCount(Candidate))] = PStart;
				++num;
			}
		}

	if (num == 0)
		foreach AllActors(class'PlayerStart', PStart)
		{
			if (PStart.bSinglePlayerStart && PStart.bEnabled && !PStart.Region.Zone.bWaterZone && PStart.Region.ZoneNumber > 0)
			{
				if (num < ArrayCount(Candidate))
					Candidate[num] = PStart;
				else if (Rand(num) < ArrayCount(Candidate))
					Candidate[Rand(ArrayCount(Candidate))] = PStart;
				++num;
			}
		}

	if (num == 0)
		foreach AllActors(class'PlayerStart', PStart)
		{
			if (PStart.bSinglePlayerStart && PStart.bEnabled && PStart.Region.ZoneNumber > 0)
			{
				if (num < ArrayCount(Candidate))
					Candidate[num] = PStart;
				else if (Rand(num) < ArrayCount(Candidate))
					Candidate[Rand(ArrayCount(Candidate))] = PStart;
				++num;
			}
		}

	if (num == 0)
		foreach AllActors(class'PlayerStart', PStart)
		{
			if ((PStart.bSinglePlayerStart || PStart.bCoopStart) && !PStart.Region.Zone.bWaterZone && PStart.Region.ZoneNumber > 0)
			{
				if (num < ArrayCount(Candidate))
					Candidate[num] = PStart;
				else if (Rand(num) < ArrayCount(Candidate))
					Candidate[Rand(ArrayCount(Candidate))] = PStart;
				++num;
			}
		}

	if (num == 0)
		foreach AllActors(class'PlayerStart', PStart)
		{
			if ((PStart.bSinglePlayerStart || PStart.bCoopStart) && PStart.Region.ZoneNumber > 0)
			{
				if (num < ArrayCount(Candidate))
					Candidate[num] = PStart;
				else if (Rand(num) < ArrayCount(Candidate))
					Candidate[Rand(ArrayCount(Candidate))] = PStart;
				++num;
			}
		}

	if (num == 0)
	{
		if (IncomingName != "")
		{
			foreach AllActors(class'Teleporter', Telep)
				if (string(Telep.Tag) ~= IncomingName && Telep.Region.ZoneNumber > 0)
					return Telep;
		}
		foreach AllActors(class'PlayerStart', PStart)
		{
			if (PStart.Region.ZoneNumber > 0)
			{
				if (num < ArrayCount(Candidate))
					Candidate[num] = PStart;
				else if (Rand(num) < ArrayCount(Candidate))
					Candidate[Rand(ArrayCount(Candidate))] = PStart;
				++num;
			}
		}
		if (num == 0)
		{
			foreach AllActors(class'Teleporter', Telep)
			{
				if (Telep.URL != "" || Telep.Region.ZoneNumber <= 0)
					continue;
				if (Telep.Tag == '')
					return Telep;
				TelepTag = string(Telep.Tag);

				num = 1;
				foreach AllActors(class'Teleporter', OtherTelep)
					if (TelepTag ~= OtherTelep.URL)
					{
						num = 0;
						break;
					}
				if (num == 1)
					return Telep;
			}
			foreach AllActors(class'NavigationPoint', Best)
			{
				if (Best.Region.ZoneNumber > 0)
					return Best;
			}
			foreach AllActors(class'PlayerStart', PStart)
				return PStart;
			foreach AllActors(class'NavigationPoint', Best)
				return Best;
			return none;
		}
	}

	if (num > ArrayCount(Candidate))
		num = ArrayCount(Candidate);

	for (i = 0; i < num; ++i)
		Score[i] = 4000 * FRand(); // randomize

	foreach AllActors(class'Pawn', OtherPlayer)
	{
		if (GameControl.GetPlayerInfo(OtherPlayer) != none)
		{
			for (i = 0; i < num; i++)
			{
				NextDist = VSize(OtherPlayer.Location - Candidate[i].Location);
				Score[i] += NextDist;
				if (NextDist < OtherPlayer.CollisionRadius + OtherPlayer.CollisionHeight)
					Score[i] -= 1000000.0;
			}
		}
	}

	BestScore = Score[0];
	Best = Candidate[0];
	for (i = 1; i < num; ++i)
	{
		if (Score[i] > BestScore)
		{
			BestScore = Score[i];
			Best = Candidate[i];
		}
	}

	return Best;
}

function bool CanSpectate(Pawn Viewer, Actor ViewTarget)
{
	if (!GameControl.IsPlayer(Viewer))
		return true;
	if (GameControl.IsParticipant(Pawn(ViewTarget)))
		return true;
	return bPlayersCanViewAnyActors;
}

function bool IsRelevant(Actor A)
{
	if (PlayerPawn(A) != none)
		Pawn(A).PlayerReplicationInfoClass = class'UGoldPlayerReplicationInfo';
	if (NaliFruit(A) != none && Level.TimeSeconds > 0)
		HandleAddedNaliFruit(NaliFruit(A));

	return Super.IsRelevant(A);
}

function Tick(float DeltaTime)
{
	local PlayerPawn Player;

	if (bRetainInventory)
	{
		foreach AllActors(class'PlayerPawn', Player)
			AdjustDispersionPistolCharge(Player);
	}

	Super.Tick(DeltaTime);
}

function AdjustDispersionPistolCharge(Pawn Player)
{
	if (Player.Weapon != none &&
		Player.Weapon.class == class'UnrealShare.DispersionPistol' &&
		Player.Weapon.AmmoType != none &&
		Player.Weapon.AmmoType.AmmoAmount < 1 &&
		!Player.Weapon.IsInState('AltFiring'))
	{
		Player.Weapon.AmmoType.AmmoAmount = 1;
	}
}

event AcceptInventory(pawn PlayerPawn)
{
	//default accept all inventory except default weapon (spawned explicitly)
	// Initialize the inventory.
	AddDefaultInventory(PlayerPawn);

	log("All inventory from" @ GameControl.UTF8_Msg(PlayerPawn.PlayerReplicationInfo.PlayerName) @ "is accepted" , 'ServerLog');
}

function AddDefaultInventory(Pawn P)
{
	if (bRetainInventory)
		ReturnSavedInventory(P);
	GameControl.AddDefaultInventory(P);
	GameControl.AddDefaultCoopInventory(P);
}

function ReturnSavedInventory(Pawn P)
{
	local UGoldPlayerReplicationInfo PlayerInfo;
	local Inventory Inv;

	PlayerInfo = UGoldPlayerReplicationInfo(P.PlayerReplicationInfo);
	if (PlayerInfo != none && PlayerInfo.Inventory != none && P.Inventory != PlayerInfo.Inventory)
	{
		for (Inv = P.Inventory; Inv != none; Inv = Inv.Inventory)
			Inv.Destroy();

		P.Inventory = PlayerInfo.Inventory;
		if (P.SelectedItem == none)
			P.SelectedItem = PlayerInfo.SelectedItem;
		if (P.Weapon == none && PlayerInfo.SelectedWeapon != none)
			PlayerInfo.SelectedWeapon.WeaponSet(P);
		PlayerInfo.Inventory = none;
	}
}

function StartPlayer(PlayerPawn Player)
{
	if (Player.PlayerReStartState == '' || Player.PlayerReStartState == 'PlayerWalking')
	{
		if (Player.Region.Zone.bWaterZone)
		{
			Player.SetPhysics(PHYS_SWIMMING);
			Player.GotoState('PlayerSwimming');
		}
		else
			Player.GotoState('PlayerWalking');
	}
	else
		Player.GotoState(Player.PlayerReStartState);
}

function bool PlayerDoesNotDamage(Pawn DamageInstigator, Pawn DamageTarget)
{
	local UGoldPlayerReplicationInfo InstigatorPlayerInfo, TargetPlayerInfo;

	if (DamageTarget == DamageInstigator)
		return false;

	InstigatorPlayerInfo = GameControl.GetPlayerInfo(DamageInstigator);
	if (InstigatorPlayerInfo == none)
		return false;
	TargetPlayerInfo = GameControl.GetPlayerInfo(DamageTarget);
	if (TargetPlayerInfo == none)
		return false;

	if (InstigatorPlayerInfo.HostileStatus != 0)
		return true;

	if (bNoFriendlyFire && TargetPlayerInfo.HostileStatus == 0)
		return true;

	return false;
}

function bool HasNoDamageImpactOn(Pawn InstigatorPawn, Pawn AffectedPawn)
{
	local UGoldPlayerReplicationInfo InstigatorPlayerInfo, AffectedPlayerInfo;

	if (InstigatorPawn == AffectedPawn)
		return false;

	InstigatorPlayerInfo = GameControl.GetPlayerInfo(InstigatorPawn);
	if (InstigatorPlayerInfo == none)
		return false;
	AffectedPlayerInfo = GameControl.GetPlayerInfo(AffectedPawn);
	if (AffectedPlayerInfo == none)
		return false;

	if (InstigatorPlayerInfo.HostileStatus == 2 && AffectedPlayerInfo.HostileStatus == 0)
		return true;

	if (!bNoFriendlyFire || !bCustomFriendlyFireImpact && !bNoFriendlyFireImpact)
		return false;

	if (InstigatorPlayerInfo.HostileStatus == 0 && AffectedPlayerInfo.HostileStatus != 0)
		return false;

	if (bNoFriendlyFireImpact)
		return true;

	return InstigatorPlayerInfo.bNoFriendlyFireImpact || AffectedPlayerInfo.bNoFriendlyFireImpact;
}

function int ReduceDamage(int Damage, name DamageType, pawn injured, pawn instigatedBy)
{
	local UGoldPlayerReplicationInfo InstigatorPlayerInfo, InjuredPlayerInfo;

	InstigatorPlayerInfo = GameControl.GetPlayerInfo(instigatedBy);
	InjuredPlayerInfo = GameControl.GetPlayerInfo(injured);

	if (HasNoDamageImpactOn(instigatedBy, injured) && (InstigatorPlayerInfo == none || InstigatorPlayerInfo.HostileStatus == 0))
		return 0;

	if (bShowDamageFromPlayers && injured != instigatedBy && PlayerPawn(injured) != none && PlayerPawn(instigatedBy) != none)
	{
		if (InjuredPlayerInfo != none && InjuredPlayerInfo.bShowDamageFromPlayers)
		{
			if (Level.TimeSeconds - InjuredPlayerInfo.LastIndicatedDamageTime >= 0.5 ||
				instigatedBy != InjuredPlayerInfo.LastIndicatedDamageInstigator)
			{
				injured.ClientMessage("\"" $ DamageType $ "\"" @ "damage by:" @ GetHumanNameOf(instigatedBy));
				InjuredPlayerInfo.LastIndicatedDamageInstigator = instigatedBy;
				InjuredPlayerInfo.LastIndicatedDamageTime = Level.TimeSeconds;
			}
		}
	}

	if (PlayerDoesNotDamage(instigatedBy, injured))
		return 0;

	if (PlayerPawn(injured) != None && (GameControl.GameEndTime > 0 || Level.NextURL != "")) // level is ended
		return 0;

	if ( (DamageType == 'Fell') && bSpecialFallDamage )
		return Min(Damage, 5);

	if (injured.Region.Zone.bNeutralZone)
		return 0;

	if ( instigatedBy == None)
		return Damage;
	//skill level modification
	if ( instigatedBy.bIsPlayer )
	{
		if ( injured == instigatedby )
		{ 
			if ( instigatedby.skill == 0 )
				Damage = 0.25 * Damage;
			else if ( instigatedby.skill == 1 )
				Damage = 0.5 * Damage;
		}
		else if ( !injured.bIsPlayer )
			Damage = float(Damage) * (1.1 - 0.1 * injured.skill);
	}
	else if ( injured.bIsPlayer )
		Damage = Damage * (0.4 + 0.2 * instigatedBy.skill);
	return (Damage * instigatedBy.DamageScaling);
}

function Killed(pawn killer, pawn Other, name damageType)
{
	local string message;
	local string KillerWeaponName;
	local string VictimWeaponName;
	local string DeathMessage;
	local UGoldPlayerReplicationInfo KillerInfo, VictimInfo;
	local bool bNoMessage;

	KillerInfo = GameControl.GetPlayerInfo(Killer);
	VictimInfo = GameControl.GetPlayerInfo(Other);

	if (VictimInfo != none)
	{
		if ( (Killer == Other) || (Killer == None) )
		{
			// suicide
			if ( LocalLog != None )
				LocalLog.LogSuicide(Other, damageType);
			if ( WorldLog != None )
				WorldLog.LogSuicide(Other, damageType);

			if (VictimInfo != none)
			{
				if (damageType == 'suicided')
				{
					if (VictimInfo.bMuted || !GameControl.CheckOptionalMessageTimeIntervals(VictimInfo, VictimInfo.LastSuicideMessageTime))
						bNoMessage = true;
				}
				else if (VictimInfo.bMuted)
				{
					if (!GameControl.CheckOptionalMessageTimeIntervals(VictimInfo, VictimInfo.LastSuicideMessageTime))
						bNoMessage = true;
				}
			}
			message = KillMessage(damageType, Other);
			if (bNoMessage)
				Other.ClientMessage("You" $ message);
			else
				BroadcastMessage(GetHumanNameOf(Other) $ message, false, 'DeathMessage');
		}
		else
		{
			if (Killer.Weapon != None)
			{
				KillerWeaponName = Killer.Weapon.ItemName;
				DeathMessage = Killer.Weapon.DeathMessage;
			}
			if (Other.Weapon != None)
				VictimWeaponName = Other.Weapon.ItemName;

			if (Other.PlayerReplicationInfo != None && Killer.PlayerReplicationInfo != None)
			{
				if ( LocalLog != None )
					LocalLog.LogKill(
							Killer.PlayerReplicationInfo.PlayerID,
							Other.PlayerReplicationInfo.PlayerID,
							KillerWeaponName,
							VictimWeaponName,
							damageType
					);
				if ( WorldLog != None )
					WorldLog.LogKill(
							Killer.PlayerReplicationInfo.PlayerID,
							Other.PlayerReplicationInfo.PlayerID,
							KillerWeaponName,
							VictimWeaponName,
							damageType
					);
			}

			if ( (DamageType == 'SpecialDamage') && (SpecialDamageString != "") )
			{
				message = ParseKillMessage(
					GetHumanNameOf(Killer),
					GetHumanNameOf(Other),
					KillerWeaponName,
					SpecialDamageString
				);
				BroadcastMessage(message, false, 'DeathMessage');
			}
			else if ( bClassicDeathmessages || (Killer.Weapon == None)
				|| ((DamageType != Killer.Weapon.MyDamageType) && (DamageType != Killer.Weapon.AltDamageType)) )
			{
				if (killer.IsA('Bots'))
					message = PlayerKillMessage(damageType, Other) $ GetHumanNameOf(killer);
				else if (ScriptedPawn(killer) != none)
				{
					message = Level.Game.CreatureKillMessage(damageType, Other);
					if (killer.MenuName != "")
						message = MergeStrings(message, killer.NameArticle) $ killer.MenuName;
					else
						message = MergeStrings(message, killer.class.name);
				}
				else
					message = killer.KillMessage(damageType, Other);
				BroadcastMessage(GetHumanNameOf(Other) $ message, false, 'DeathMessage');
			} 
			else 
			{
				message = ParseKillMessage(
					GetHumanNameOf(Killer),
					GetHumanNameOf(Other),
					KillerWeaponName,
					DeathMessage
				);
				BroadcastMessage(message, false, 'DeathMessage');
			}
		}
	}
	else if (bShowPlayerKills && PlayerPawn(killer) != none &&
		KillerInfo != none && KillerInfo.bShowPlayerKills)
	{
		message = GetHumanNameOf(Other);
		if (Left(message, 1) == " ")
			message = "You killed" $ message;
		else
			message = "You killed" @ message;
		killer.ClientMessage(message);
	}

	if (Other != None)
		ScoreKill(killer, Other);
	if (GameControl.GetPlayerInfo(Other) != none)
		GameControl.GetPlayerInfo(Other).RespawnDelay = FMax(0.5, PlayerRespawnDelay);
}

function ScoreKill(pawn Killer, pawn Victim)
{
	local int Attitude;

	Attitude = TentativeAttitudeToPlayer(Victim, PlayerPawn(Killer));
	GameControl.ScoreKill(Killer, Victim, Attitude);
	if (Attitude < 0 && UGoldCoopReplicationInfo(GameReplicationInfo) != none)
		++UGoldCoopReplicationInfo(GameReplicationInfo).CoopKillsCount;
}

static function string MergeStrings(coerce string S1, coerce string S2)
{
	return class'UGoldMP'.static.MergeStrings(S1, S2);
}

function int TentativeAttitudeToPlayer(Pawn P, Playerpawn Player)
{
	if (PlayerPawn(P) != none)
	{
		if (GameControl.GetPlayerInfo(P) == none)
			return 0;
		if (GameControl.GetPlayerInfo(P).HostileStatus != 0)
			return -1;
		return 1;
	}
	if (Bots(P) == none && ScriptedPawn(P) == none)
		return 0;
	if (Nali(P) != none)
		return 1;
	if (Cow(P) != none)
		return 0;
	if (P.IsA('SpaceMarine'))
		return -1;
	if (Player != none)
	{
		if (P.AttitudeToPlayer >= ATTITUDE_Friendly ||
			Bots(P) != none && Bots(P).AttitudeTo(Player) >= ATTITUDE_Friendly ||
			ScriptedPawn(P) != none && ScriptedPawn(P).AttitudeTo(Player) >= ATTITUDE_Friendly)
		{
			return 1;
		}
		if (P.default.AttitudeToPlayer < ATTITUDE_Ignore ||
			P.AttitudeToPlayer < ATTITUDE_Ignore ||
			Bots(P) != none && Bots(P).AttitudeTo(Player) < ATTITUDE_Ignore ||
			ScriptedPawn(P) != none && ScriptedPawn(P).AttitudeTo(Player) < ATTITUDE_Ignore)
		{
			return -1;
		}
		return 0;
	}

	if (P.AttitudeToPlayer >= ATTITUDE_Friendly)
		return 1;
	if (P.default.AttitudeToPlayer < ATTITUDE_Ignore || P.AttitudeToPlayer < ATTITUDE_Ignore)
		return -1;
	return 0;
}

function string PlayerKillMessage(name damageType, pawn Other)
{
	return GameControl.PlayerKillMessage(damageType, Other);
}

function string GetHumanNameOf(Pawn P)
{
	return GameControl.GetHumanNameOf(P);
}

function DiscardInventory(Pawn P)
{
	local Inventory inv, next_inv;
	local UGoldPlayerReplicationInfo PlayerInfo;

	PlayerInfo = UGoldPlayerReplicationInfo(P.PlayerReplicationInfo);

	if (PlayerPawn(P) != none &&
		PlayerInfo != none &&
		PlayerInfo.HostileStatus == 0 &&
		bRetainInventory &&
		PlayerInfo.Inventory == none)
	{
		for (inv = P.Inventory; inv != none; inv = next_inv)
		{
			next_inv = inv.Inventory;

			if (inv.bIsAnArmor)
				inv.Destroy();
			else if (inv.bActivatable && inv.IsInState('Activated'))
				inv.Activate();
		}

		if (P.Weapon != none)
		{
			P.Weapon.bChangeWeapon = true;
			P.Weapon.Finish();
		}

		PlayerInfo.SelectedWeapon = P.Weapon;
		PlayerInfo.SelectedItem = P.SelectedItem;
		PlayerInfo.Inventory = P.Inventory;

		P.Weapon = none;
		P.SelectedItem = none;
		P.Inventory = none;

		GameControl.MakePlayerController(P);

		return;
	}
	Super.DiscardInventory(P);
	GameControl.MakePlayerController(P);
}

function bool RestartPlayer(Pawn Player)
{
	return GameControl.CoopCanRestartPlayer(Player) && super.RestartPlayer(Player);
}

function SendPlayer(PlayerPawn aPlayer, string URL)
{
	local UGoldPlayerReplicationInfo PlayerInfo;

	if (GameControl.GameEndTime > 0 || Level.NextURL != "")
		return;

	if (GameControl.bEndingDisallowed)
	{
		aPlayer.ClientMessage("Ending is disallowed");
		foreach AllActors(class'UGoldPlayerReplicationInfo', PlayerInfo)
		{
			if (PlayerInfo.Player() != aPlayer && PlayerInfo.GetAdminLevel() > 0)
				PlayerInfo.Player().ClientMessage(GetHumanNameOf(aPlayer) @ "is reaching the end. Ending is disallowed.");
		}
		return;
	}

	PlayerInfo = UGoldPlayerReplicationInfo(aPlayer.PlayerReplicationInfo);

	if (PlayerInfo == none)
		return;

	GameControl.UpdateVoteEndStatistics(PlayerInfo, true);
	if (!GameControl.bEndingEnabled || !PlayerInfo.VotedToEnd() || PlayerInfo.HostileStatus != 0)
		return;
	GameControl.PlayerWhoEndedLevel = aPlayer;
	BroadcastMessage(PlayerInfo.PlayerName $ " ended the level");
	EndGame(PlayerInfo.PlayerName $ " ended the level");
	GameControl.NextURL = URL;
}

function ChangeName(Pawn Player, string NewName, bool bNameChange)
{
	GameControl.ChangeName(Player, NewName, bNameChange);
}

function AdjustEndTeleporters()
{
	local Teleporter telep;

	if (bShowEnds)
	{
		foreach AllActors(class'Teleporter', telep)
			if (InStr(telep.URL, "/") >= 0 || InStr(telep.URL, "#") >= 0)
			{
				telep.bHidden = false;
				telep.DrawType = DT_Sprite;
				telep.DrawScale = FClamp(telep.DrawScale, 1, 10);
				telep.Texture = telep.default.Texture;
				telep.bAlwaysRelevant = true;
			}
	}
}

function bool PickupQuery(Pawn P, Inventory Item)
{
	local UGoldPlayerReplicationInfo PlayerInfo;

	PlayerInfo = GameControl.GetPlayerInfo(P);
	if (PlayerInfo == none)
		return super.PickupQuery(P, Item);
	if (PlayerInfo.HostileStatus > 0 && (bNoInventoryRespawns || Weapon(Item) == none))
		return false;
	return super.PickupQuery(P, Item);
}

function bool ShouldRespawn(Actor A)
{
	if (Inventory(A) == none || bNoInventoryRespawns)
		return false;
	if (Weapon(A) != none)
	{
		if (WeaponRespawnTime == -1)
			return Weapon(A).RespawnTime > 0 && !Weapon(A).bHeldItem;
		if (WeaponRespawnTime > 0)
		{
			if (Weapon(A).bHeldItem || Weapon(A).RespawnTime <= 0)
				return false;
			Weapon(A).RespawnTime = WeaponRespawnTime;
			return true;
		}
	}
	if (Pickup(A) != none)
	{
		if (PickupRespawnTime == -1)
			return bRespawnPickups && Pickup(A).RespawnTime > 0 && !Pickup(A).bHeldItem;
		if (PickupRespawnTime > 0)
		{
			if (!bRespawnPickups || Pickup(A).bHeldItem)
				return false;
			Pickup(A).RespawnTime = PickupRespawnTime;
			return true;
		}
	}
	return super.ShouldRespawn(A) || (bRespawnPickups && Pickup(A) != none && Pickup(A).RespawnTime > 0);
}

function bool SetPause(bool bPause, PlayerPawn P)
{
	return GameControl.SetPause(bPause, P);
}

function bool AllowsBroadcast(Actor Broadcaster, int Len)
{
	SentText += Len;

	return SentText <= 512 || Level.Pauser != "";
}

// don't let players restart map
function RestartGame() {}

function SwitchToNextLevel()
{
	Level.ServerTravel(GameControl.NextURL, true);
}

function EndGame(string Reason)
{
	local UGoldPlayerReplicationInfo PlayerInfo;
	local UGoldMPPlayerPickup P;

	LogS("EndGame:" @ Reason);
	GameControl.EndGame();
	GameControl.SetGameStatusStr(" ");
	GameControl.NextURL = GameControl.NextMapURL;

	foreach AllActors(class'UGoldPlayerReplicationInfo', PlayerInfo)
	{
		if (PlayerInfo.bMoveCameraOnLevelEnd)
		{
			if (PlayerInfo.Player() != GameControl.PlayerWhoEndedLevel)
			{
				PlayerInfo.Player().GotoState('GameEnded');
				PlayerInfo.Player().ClientGameEnded();
				PlayerInfo.Player().ViewTarget = GameControl.PlayerWhoEndedLevel;
			}
			PlayerInfo.Player().bBehindView = true;
		}
	}

	foreach AllActors(class'UGoldMPPlayerPickup', P)
		P.ShowScoreBoard();
}


function UpdateSavedScores(UGoldPlayerReplicationInfo P)
{
	local UGoldPlayerReplicationInfo PlayerInfo;
	local int i;

	if (P == none)
		return;
	foreach AllActors(class'UGoldPlayerReplicationInfo', PlayerInfo)
		if (P != PlayerInfo && PlayerInfo.PlayerName ~= P.PlayerName && PlayerInfo.IPAddress == P.IPAddress)
		{
			PlayerInfo.Score += P.Score;
			PlayerInfo.Deaths += P.Deaths;
			return;
		}

	for (i = 0; i < ArrayCount(SavedScores); ++i)
		if (SavedScores[i].PlayerName == "")
			break;
	if (i == ArrayCount(SavedScores))
		i = 0;

	SavedScores[i].PlayerName = P.PlayerName;
	SavedScores[i].IPAddress = P.IPAddress;
	SavedScores[i].Score = P.Score;
	SavedScores[i].Deaths = P.Deaths;
}

function ObtainSavedScores(UGoldPlayerReplicationInfo P)
{
	local int i;

	if (P == none)
		return;

	for (i = 0; i < ArrayCount(SavedScores); ++i)
		if (SavedScores[i].PlayerName ~= P.PlayerName && SavedScores[i].IPAddress == P.IPAddress)
		{
			P.Score = SavedScores[i].Score;
			P.Deaths = SavedScores[i].Deaths;
			while (++i < ArrayCount(SavedScores))
				SavedScores[i - 1] = SavedScores[i];
			SavedScores[i - 1].PlayerName = "";
			return;
		}
}

function HandleAddedNaliFruit(NaliFruit A)
{
	local NaliFruit OtherFruit;
	local int i;
	if (MaxAddedNaliFruits > 0)
	{
		foreach AllActors(class'NaliFruit', OtherFruit)
			if (OtherFruit.bHeldItem)
			{
				++i;
			}
		if (i >= MaxAddedNaliFruits)
		{
			foreach AllActors(class'NaliFruit', OtherFruit)
				if (OtherFruit.bHeldItem)
				{
					OtherFruit.Destroy();
					if (--i < MaxAddedNaliFruits)
						break;
				}
		}
	}
	A.bHeldItem = true;
}

function LogS(coerce string S)
{
	GameControl.LogS(S);
}


exec function RejectNewConnections(optional int Minutes)
{
	GameControl.AdminControl.ServerRejectNewConnections(Minutes);
}

exec function EnableNewConnections()
{
	GameControl.AdminControl.ServerEnableNewConnections();
}

exec function NextLevel()
{
	GameControl.SwitchToNextMap();
}

exec function GotoLevel(string URL)
{
	Level.ServerTravel(URL, true);
}

defaultproperties
{
				bAllowNoBlock=True
				bAllowVoteSkip=True
				bKeepScores=True
				bNoTeamKills=True
				bPlayersCanViewAnyActors=True
				bRetainInventory=True
				bShowEnds=True
				ExtraServerTravelPause=2.000000
				IdleTimeThreshold=1
				MaxAddedNaliFruits=200
				PlayerRespawnDelay=1.000000
				VoteEndMinPct=51
				VoteSkipMinPct=51
				VoteSkipMinPlayers=3
				ScoreBoardType=Class'KOS.UGoldMPScoreBoard'
				HUDType=Class'KOS.UGoldCoopHUD'
				GameReplicationInfoClass=Class'KOS.UGoldCoopReplicationInfo'
}
